from ._box_plot import BoxPlot
from ._styled_box_plot import StyledBoxPlot
from ._foreground_scatter import ForegroundScatter

from ._styled_box_plot import StyledBoxPlot, boxplot

__all__ = ['StyledBoxPlot', 'boxplot']